"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const getSchemaByType_1 = require("../util/getSchemaByType");
const convert_1 = require("../util/convert");
const util_1 = require("../util");
const util_2 = require("../util");
const util_3 = require("../util");
const METADATA = {};
console.log('[egg-controller] load transformer: response-schema.');
function transformer(program) {
    const typeChecker = program.getTypeChecker();
    return {
        before(ctx) {
            return (sourceFile) => {
                return util_1.walker(sourceFile, ctx, before(ctx, sourceFile, typeChecker));
            };
        },
        after(ctx) {
            return (sourceFile) => {
                return util_1.walker(sourceFile, ctx, after(ctx, sourceFile, typeChecker));
            };
        },
    };
}
exports.default = transformer;
function before(_, sourceFile, typeChecker) {
    // 当前编译文件需要保存的数据
    const fileData = (METADATA[sourceFile.fileName] = {
        // 方法定义
        methodDefine: {},
        schemaObjects: {},
        typeCache: [],
    });
    return (node) => {
        if (!ts.isDecorator(node) ||
            util_1.getDecoratorName(node) !== 'route' ||
            !ts.isCallExpression(node.expression)) {
            return node;
        }
        const decoratorFactory = node.expression;
        const methodNode = node.parent;
        const clsNode = node.parent.parent;
        if (!ts.isClassDeclaration(clsNode) || !ts.isMethodDeclaration(methodNode)) {
            return node;
        }
        const clsName = clsNode.name.text;
        const clsMethod = util_1.getClsMethodKey(clsNode.name.text, methodNode.name.getText());
        const lastArg = decoratorFactory.arguments.length &&
            decoratorFactory.arguments[decoratorFactory.arguments.length - 1];
        const oldConfigArg = lastArg && ts.isObjectLiteralExpression(lastArg) && lastArg;
        const oldSchemasProp = util_3.getField(oldConfigArg, 'schemas');
        const oldSchemasPropObj = oldSchemasProp && oldSchemasProp.initializer;
        const config = {
            typeChecker,
            schemaObjects: fileData.schemaObjects,
            typeCache: fileData.typeCache,
        };
        const type = typeChecker.getTypeAtLocation(methodNode);
        const callSignatures = type.getCallSignatures()[0];
        const appendProps = [];
        // params parser
        if (!util_2.hasField(oldSchemasPropObj, 'requestBody')) {
            try {
                const parameters = callSignatures.getParameters();
                const paramSchema = parameters.reduce((s, p) => {
                    const paramType = typeChecker.getTypeAtLocation(p.valueDeclaration);
                    s[`${p.escapedName}`] = getSchemaByType_1.getSchemaByType(paramType, config);
                    return s;
                }, {});
                appendProps.push(ts.createPropertyAssignment('requestBody', convert_1.convert({ type: 'object', properties: paramSchema })));
            }
            catch (error) {
                console.log('warn:', `parse ${clsName}.${clsMethod} params fail!`, error);
            }
        }
        // returnType parser
        if (!util_2.hasField(oldSchemasPropObj, 'response')) {
            try {
                const returnType = typeChecker.getReturnTypeOfSignature(callSignatures);
                let responseSchema;
                if (util_1.getValue(() => returnType.symbol.escapedName) === 'Promise') {
                    responseSchema = getSchemaByType_1.getSchemaByType(returnType.typeArguments[0], {
                        typeChecker,
                        schemaObjects: fileData.schemaObjects,
                        typeCache: fileData.typeCache,
                    });
                }
                else {
                    responseSchema = getSchemaByType_1.getSchemaByType(returnType, config);
                }
                appendProps.push(ts.createPropertyAssignment('response', convert_1.convert(responseSchema)));
            }
            catch (error) {
                console.log('warn:', `parse ${clsName}.${clsMethod} returnType fail!`, error);
            }
        }
        // reference type define
        if (!util_2.hasField(oldSchemasPropObj, 'components')) {
            appendProps.push(ts.createPropertyAssignment('components', ts.createIdentifier('__SchemaDefinition')));
        }
        const schemasProp = oldSchemasProp
            ? ts.updatePropertyAssignment(oldSchemasProp, ts.createIdentifier('schemas'), ts.createObjectLiteral([
                ...(ts.isObjectLiteralExpression(oldSchemasProp.initializer)
                    ? oldSchemasProp.initializer.properties
                    : []),
                ...appendProps,
            ], false))
            : ts.createPropertyAssignment('schemas', ts.createObjectLiteral(appendProps, false));
        return ts.updateDecorator(node, ts.updateCall(decoratorFactory, decoratorFactory.expression, decoratorFactory.typeArguments, [
            ...decoratorFactory.arguments.filter(arg => arg !== oldConfigArg),
            oldConfigArg
                ? ts.updateObjectLiteral(oldConfigArg, [...oldConfigArg.properties, schemasProp])
                : ts.createObjectLiteral([schemasProp], false),
        ]));
    };
}
exports.before = before;
function after(_, sourceFile, __) {
    const fileData = METADATA[sourceFile.fileName];
    if (fileData) {
        sourceFile.statements = ts.createNodeArray([
            ts.createVariableStatement([], ts.createVariableDeclarationList([
                ts.createVariableDeclaration('__SchemaDefinition', undefined, convert_1.convert(fileData.schemaObjects)),
            ], ts.NodeFlags.Const)),
            ...sourceFile.statements,
        ]);
    }
    return node => node;
}
exports.after = after;
